from typing import TypeVar, Generic, overload, Union, Optional, List, Tuple
from collections.abc import Callable
from time import sleep

from mafunca.triple import Left, Nothing, TUtils
from mafunca.common.resilient_support import Uncaught, Report
from mafunca.common.exceptions import MonadError
import mafunca.common.panics as panics


__all__ = ['of', 'unit', 'ResilientSyncPrime', 'ResilientSyncCont', 'insist']


Exc = TypeVar('Exc', bound=Exception)
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
L = TypeVar('L')


_ORIGIN_LINK = "__mafunca_resilient_origin__"


def _continuer(fn: Callable, bad_evaluator: Callable[..., bool]) -> Callable:
    """
        Special sync closure for continuation through the standard chain method.
        With short circuits on 'Uncaught' and custom 'bad' entities.
    """
    def _continuer_inner(arg):
        if isinstance(arg, Uncaught) or bad_evaluator(arg):
            return arg
        return fn(arg)

    setattr(_continuer_inner, _ORIGIN_LINK, fn)
    return _continuer_inner


def _catcher(fn: Callable) -> Callable:
    """Special sync closure for catching errors"""
    def _catcher_inner(arg):
        if isinstance(arg, Uncaught):
            return fn(arg.error)
        return arg

    setattr(_catcher_inner, _ORIGIN_LINK, fn)
    return _catcher_inner


def _ensurer(fn: Callable) -> Callable:
    """A sync closure simulating finally"""
    def _ensurer_inner(arg):
        fn()
        return arg

    setattr(_ensurer_inner, _ORIGIN_LINK, fn)
    return _ensurer_inner


def _extract_from_closure(closure: Callable) -> Callable:
    """Extract origin function from closures like '_continuer' and etc"""
    origin = getattr(closure, _ORIGIN_LINK, None)
    if origin is None:
        return closure
    return origin


class _ResilientSync(Generic[A]):
    __slots__ = ('_effect',)

    def __init__(self, effect: Callable[..., A]):
        panics.on_coroutine(effect, monad_name=self.__class__.__name__, method='__init__')
        self._effect = effect

    @overload
    def chain(self: '_ResilientSync[Uncaught[Exc]]', fn: Callable[[B], C]) -> 'ResilientSyncCont[Uncaught[Exc]]': pass
    @overload
    def chain(self: '_ResilientSync[Left[L]]', fn: Callable[[B], C]) -> 'ResilientSyncCont[Left[L]]': pass
    @overload
    def chain(self: '_ResilientSync[Nothing]', fn: Callable[[B], C]) -> 'ResilientSyncCont[Nothing]': pass
    @overload
    def chain(self, fn: Callable[[A], '_ResilientSync[B]']) -> 'ResilientSyncCont[B]': pass
    @overload
    def chain(self, fn: Callable[[A], B]) -> 'ResilientSyncCont[B]': pass

    def chain(self, fn):
        """
            Combines the logic of both 'map' and 'bind' in regular monads.
            :raises MonadError: panics on coroutine function
        """
        panics.on_coroutine(fn, monad_name=self.__class__.__name__, method='chain')
        return ResilientSyncCont(_continuer(fn, bad_evaluator=TUtils.is_bad), past=self)

    @overload
    def catch(self, fn: Callable[[Exc], '_ResilientSync[B]']) -> 'ResilientSyncCont[Union[A, B]]': pass
    @overload
    def catch(self, fn: Callable[[Exc], B]) -> 'ResilientSyncCont[Union[A, B]]': pass

    def catch(self, fn):
        """
            Handles errors(Exception heirs) that occurred earlier in the chain.
            :raises MonadError: panics on coroutine function
        """
        panics.on_coroutine(fn, monad_name=self.__class__.__name__, method='catch')
        return ResilientSyncCont(_catcher(fn), past=self)

    def ensure(self, fn: Callable[[], None]) -> 'ResilientSyncCont[A]':
        """
            Guaranteed to execute the function-parameter, similar to try finally.
            :raises MonadError: panics on coroutine function
        """
        panics.on_coroutine(fn, monad_name=self.__class__.__name__, method='ensure')
        return ResilientSyncCont(_ensurer(fn), past=self)


class ResilientSyncPrime(_ResilientSync[A]):
    """Use only for typing purposes. To create monads, use the 'of' and 'unit' functions and chained methods"""

    def __init__(self, effect: Callable[[], A]):
        super().__init__(effect)

    @property
    def effect(self) -> Callable[[], A]:
        return self._effect

    def run(self, rebuild: bool = False) -> Report[Union[A, Uncaught], Optional['ResilientSyncPrime']]:
        """
            Starts the chain
            :raises MonadError: panics on coroutine function
        """
        try:
            result = self._effect()
        except Exception as exc:
            result = Uncaught(exc)
        if rebuild and (isinstance(result, Uncaught) or TUtils.is_bad(result)):
            return Report(result, ResilientSyncPrime(self._effect), self._effect, None)
        return Report(result, chain_from_failure=None, faulty=None, last_success=None)


def _unwind(persist: 'ResilientSyncCont') -> Tuple['ResilientSyncPrime', List[Callable]]:
    """Unwinding the chain to the first effect"""
    current, effects = persist, []
    while isinstance(current, ResilientSyncCont):
        effects.append(current.effect)
        current = current.past
    if isinstance(current, ResilientSyncPrime):
        return current, effects
    raise MonadError(
        monad='ResilientSync',
        method='chain starter method',
        message="Violation of the usage contract - the first element in the chain must be a 'ResilientSyncPrime' entity"
    )


def _execute(value: A, fn: Callable[[A], B]) -> Union[B, Uncaught[Exc]]:
    """
        Execute function from chain with catching errors.
        MonadError is not suppressed.
        :raises MonadError: panics on coroutine function
    """
    panics.on_coroutine(fn, monad_name='ResilientSync', method='run')
    try:
        return fn(value)
    except Exception as exc:
        if isinstance(exc, MonadError):
            raise exc
        return Uncaught(exc)


def _make_effect(val: A) -> Callable[[], A]:
    """To prevent lambda from updating the reference when it is reassigned  in a loop"""
    return lambda: val


_SyncSub = TypeVar('_SyncSub', bound=_ResilientSync)


class ResilientSyncCont(_ResilientSync[B]):
    """Use only for typing purposes. To create monads, use the 'of' and 'unit' functions and chained methods"""

    __slots__ = ('_past',)

    def __init__(self, effect: Callable[[A], B], past: _SyncSub):
        super().__init__(effect)
        self._past = past

    @property
    def effect(self) -> Callable[[A], B]:
        return self._effect

    @property
    def past(self) -> _SyncSub:
        return self._past

    def run(self, rebuild: bool = False) -> Report[Union[B, Uncaught], Optional['ResilientSyncCont']]:
        """
            Starts the chain
            :raises MonadError: panics on coroutine function
        """
        persist_prime, cons = _unwind(persist=self)
        report_prime = persist_prime.run(rebuild=rebuild)
        result, restored = report_prime.result, report_prime.chain_from_failure
        faulty, last_success = report_prime.faulty, report_prime.last_success

        for i in range(len(cons) - 1, -1, -1):
            result_new, restored_new = _execute(result, cons[i]), None
            faulty_new, last_success_new = None, None
            if isinstance(result_new, (ResilientSyncPrime, ResilientSyncCont)):
                rpt = result_new.run(rebuild=rebuild)
                result_new, restored_new = rpt.result, rpt.chain_from_failure
                faulty_new, last_success_new = rpt.faulty, rpt.last_success

            if rebuild:
                if restored_new:
                    # if we're in this branch, it's the first failure!
                    restored, faulty, last_success = restored_new, faulty_new, last_success_new
                elif isinstance(result_new, Uncaught) or TUtils.is_bad(result_new):
                    if restored is None:
                        prime = ResilientSyncPrime(_make_effect(result))
                        restored, faulty = ResilientSyncCont(cons[i], past=prime), _extract_from_closure(cons[i])
                        last_success = result
                    else:
                        restored = ResilientSyncCont(cons[i], past=restored)
                else:
                    restored, faulty, last_success = None, None, result_new
            result = result_new

        return Report(result, chain_from_failure=restored, faulty=faulty, last_success=last_success)


def of(value: A) -> ResilientSyncPrime[A]:
    """
        Lazy monad for resilient SYNC ONLY effects.
        Resilient means, that if there is a failure in the chain, it can return a short chain from the point of failure.
        Automatically catches errors and wraps them in an 'Uncaught' object.
        Works with bad 'Triple' and 'Uncaught' entities using the short-circuit principle.
    """
    return ResilientSyncPrime(lambda: value)


def unit(fn: Callable[[], A]) -> ResilientSyncPrime[A]:
    """
       Lazy monad for resilient SYNC ONLY effects.
       Resilient means, that if there is a failure in the chain, it can return a short chain from the point of failure.
       Automatically catches errors and wraps them in an 'Uncaught' object.
       Works with bad 'Triple' and 'Uncaught' entities using the short-circuit principle.
    """
    return ResilientSyncPrime(fn)


def insist(
        resilient: Union[ResilientSyncPrime[A], ResilientSyncCont[A]],
        attempts: int = 1,
        pause_between: Union[int, float] = 0
) -> Report[Union[A, Uncaught[Exc]], Optional['ResilientSyncCont']]:
    """Makes 'attempts' to execute a 'resilient' chain with 'pause_between' intervals between them"""
    chain, report = resilient, Report(None, None, None, None)

    for _ in range(attempts):
        report = chain.run(rebuild=True)
        if not report.is_ok:
            chain = report.chain_from_failure
            sleep(pause_between)
            continue
        return report

    return report



from typing import TypeVar, Generic, NoReturn, Optional
from collections.abc import Callable


__all__ = ['Uncaught', 'Report']


Exc = TypeVar('Exc', bound=Exception)


class Uncaught(Generic[Exc]):
    """Special container for uncaught by user code exception"""

    __slots__ = ('_error',)

    def __init__(self, error: Exc):
        self._error = error

    @property
    def error(self) -> Exc:
        return self._error

    def throw(self) -> NoReturn:
        """:raises Exc: raises inner exception"""
        raise self._error

    def __repr__(self):
        return f'Uncaught({self._error.__class__})'


A = TypeVar('A')
B = TypeVar('B')


class Report(Generic[A, B]):
    """A container that contains the result, a shortened chain (optional) and a failure function (optional)"""

    __slots__ = ('_result', '_chain_from_failure', '_faulty', '_last_success')

    def __init__(self, result: A, chain_from_failure: B, faulty: Optional[Callable], last_success):
        self._result = result
        self._chain_from_failure = chain_from_failure
        self._faulty = faulty
        self._last_success = last_success

    @property
    def result(self) -> A:
        return self._result

    @property
    def chain_from_failure(self) -> B:
        return self._chain_from_failure

    @property
    def faulty(self) -> Optional[Callable]:
        return self._faulty

    @property
    def last_success(self):
        return self._last_success

    @property
    def is_ok(self) -> bool:
        return self._chain_from_failure is None

    def __repr__(self):
        result = f'result={self._result}'
        chain = f'chain_from_failure={self._chain_from_failure}'
        faulty = f'faulty={self._faulty}'
        last_success = f'last_success={self._last_success}'
        return f'Report({result}, {chain}, {faulty}, {last_success})'



from typing import TypeVar, Generic, overload, Union, Optional, List, Tuple
from collections.abc import Callable, Awaitable
import inspect
import asyncio

from mafunca.triple import Left, Nothing, TUtils
from mafunca.common.resilient_support import Uncaught, Report
from mafunca.common.exceptions import MonadError


__all__ = ['of', 'unit', 'ResilientPrime', 'ResilientCont', 'insist']


Exc = TypeVar('Exc', bound=Exception)
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
L = TypeVar('L')


async def _maybe_await(obj: Union[Awaitable[A], A]) -> A:
    return await obj if inspect.isawaitable(obj) else obj


_ORIGIN_LINK = "__mafunca_resilient_origin__"


def _continuer(fn: Callable, bad_evaluator: Callable[..., bool]) -> Callable[..., Awaitable]:
    """
        Special async closure for continuation through the standard chain method.
        With short circuits on 'Uncaught' and custom 'bad' entities.
    """
    async def _continuer_inner(arg):
        if isinstance(arg, Uncaught) or bad_evaluator(arg):
            return arg
        return await _maybe_await(fn(arg))

    setattr(_continuer_inner, _ORIGIN_LINK, fn)
    return _continuer_inner


def _catcher(fn: Callable) -> Callable[..., Awaitable]:
    """Special async closure for catching errors"""
    async def _catcher_inner(arg):
        if isinstance(arg, Uncaught):
            return await _maybe_await(fn(arg.error))
        return arg

    setattr(_catcher_inner, _ORIGIN_LINK, fn)
    return _catcher_inner


def _ensurer(fn: Callable) -> Callable[..., Awaitable]:
    """An async closure simulating finally"""
    async def _ensurer_inner(arg):
        await _maybe_await(fn())
        return arg

    setattr(_ensurer_inner, _ORIGIN_LINK, fn)
    return _ensurer_inner


def _extract_from_closure(closure: Callable) -> Callable:
    """Extract origin function from closures like '_continuer' and etc"""
    origin = getattr(closure, _ORIGIN_LINK, None)
    if origin is None:
        return closure
    return origin


class _Resilient(Generic[A]):
    __slots__ = ('_effect',)

    def __init__(self, effect: Callable[..., A]):
        self._effect = effect

    @overload
    def chain(self: '_Resilient[Uncaught[Exc]]', fn: Callable[[B], C]) -> 'ResilientCont[Uncaught[Exc]]': pass
    @overload
    def chain(self: '_Resilient[Left[L]]', fn: Callable[[B], C]) -> 'ResilientCont[Left[L]]': pass
    @overload
    def chain(self: '_Resilient[Nothing]', fn: Callable[[B], C]) -> 'ResilientCont[Nothing]': pass
    @overload
    def chain(self, fn: Callable[[A], '_Resilient[B]']) -> 'ResilientCont[B]': pass
    @overload
    def chain(self, fn: Callable[[A], Awaitable[B]]) -> 'ResilientCont[B]': pass
    @overload
    def chain(self, fn: Callable[[A], B]) -> 'ResilientCont[B]': pass

    def chain(self, fn):
        """Combines the logic of both 'map' and 'bind' in regular monads"""
        return ResilientCont(_continuer(fn, bad_evaluator=TUtils.is_bad), past=self)

    @overload
    def catch(self, fn: Callable[[Exc], '_Resilient[B]']) -> 'ResilientCont[Union[A, B]]': pass
    @overload
    def catch(self, fn: Callable[[Exc], Awaitable[B]]) -> 'ResilientCont[Union[A, B]]': pass
    @overload
    def catch(self, fn: Callable[[Exc], B]) -> 'ResilientCont[Union[A, B]]': pass

    def catch(self, fn):
        """Handles errors(Exception heirs) that occurred earlier in the chain."""
        return ResilientCont(_catcher(fn), past=self)

    def ensure(self, fn: Callable[[], Union[Awaitable[None], None]]) -> 'ResilientCont[A]':
        """Guaranteed to execute the function-parameter, similar to try finally."""
        return ResilientCont(_ensurer(fn), past=self)


class ResilientPrime(_Resilient[A]):
    """Use only for typing purposes. To create monads, use the 'of' and 'unit' functions and chained methods"""

    def __init__(self, effect: Callable[[], A]):
        super().__init__(effect)

    @property
    def effect(self) -> Callable[[], A]:
        return self._effect

    async def _launch(self, rebuild: bool = False) -> Report[Union[A, Uncaught[Exc]], Optional['ResilientPrime']]:
        try:
            result = await _maybe_await(self._effect())
        except Exception as exc:
            result = Uncaught(exc)
        if rebuild and (isinstance(result, Uncaught) or TUtils.is_bad(result)):
            return Report(result, ResilientPrime(self._effect), self._effect, None)
        return Report(result, chain_from_failure=None, faulty=None, last_success=None)

    async def run(
            self,
            rebuild: bool = False,
            delay: Optional[Union[int, float]] = None
    ) -> Report[Union[A, Uncaught], Optional['ResilientPrime']]:
        """
            Async - starts the chain.
            :raises TimeoutError: delay is not None and the waiting time has been exceeded.
        """
        if delay is None:
            return await self._launch(rebuild=rebuild)
        else:
            async with asyncio.timeout(delay=delay):
                return await self._launch(rebuild=rebuild)

    def to_task(self, rebuild: bool = False) -> asyncio.Task:
        """Wrap the chain starter method into a Task"""
        return asyncio.create_task(self._launch(rebuild=rebuild))


def _unwind(persist: 'ResilientCont') -> Tuple['ResilientPrime', List[Callable]]:
    """Unwinding the chain to the first effect"""
    current, effects = persist, []
    while isinstance(current, ResilientCont):
        effects.append(current.effect)
        current = current.past
    if isinstance(current, ResilientPrime):
        return current, effects
    raise MonadError(
        monad='Resilient',
        method='chain starter method',
        message="Violation of the usage contract - the first element in the chain must be a 'ResilientPrime' entity"
    )


async def _execute(value: A, fn: Callable):
    """
        Execute function from chain with catching errors.
        MonadError is not suppressed.
    """
    try:
        return await fn(value)
    except Exception as exc:
        if isinstance(exc, MonadError):
            raise exc
        return Uncaught(exc)


def _make_effect(val: A) -> Callable[[], A]:
    """To prevent lambda from updating the reference when it is reassigned  in a loop"""
    return lambda: val


_Sub = TypeVar('_Sub', bound=_Resilient)


class ResilientCont(_Resilient[B]):
    """Use only for typing purposes. To create monads, use the 'of' and 'unit' functions and chained methods"""

    __slots__ = ('_past',)

    def __init__(self, effect: Callable[[A], B], past: _Sub):
        super().__init__(effect)
        self._past = past

    @property
    def effect(self) -> Callable[[A], B]:
        return self._effect

    @property
    def past(self) -> _Sub:
        return self._past

    async def _launch(self, rebuild: bool = False) -> Report[Union[B, Uncaught[Exc]], Optional['ResilientCont']]:
        """
            Starts the chain
            :raises MonadError: violations of monadic contracts
        """
        persist_prime, cons = _unwind(persist=self)
        report_prime = await persist_prime.run(rebuild=rebuild)
        result, restored = report_prime.result, report_prime.chain_from_failure
        faulty, last_success = report_prime.faulty, report_prime.last_success

        for i in range(len(cons) - 1, -1, -1):
            result_new, restored_new = await _execute(result, cons[i]), None
            faulty_new, last_success_new = None, None
            if isinstance(result_new, (ResilientPrime, ResilientCont)):
                rpt = await result_new.run(rebuild=rebuild)
                result_new, restored_new = rpt.result, rpt.chain_from_failure
                faulty_new, last_success_new = rpt.faulty, rpt.last_success

            if rebuild:
                if restored_new:
                    # if we're in this branch, it's the first failure!
                    restored, faulty, last_success = restored_new, faulty_new, last_success_new
                elif isinstance(result_new, Uncaught) or TUtils.is_bad(result_new):
                    if restored is None:
                        prime = ResilientPrime(_make_effect(result))
                        restored, faulty = ResilientCont(cons[i], past=prime), _extract_from_closure(cons[i])
                        last_success = result
                    else:
                        restored = ResilientCont(cons[i], past=restored)
                else:
                    restored, faulty, last_success = None, None, result_new
            result = result_new

        return Report(result, chain_from_failure=restored, faulty=faulty, last_success=last_success)

    async def run(
            self,
            rebuild: bool = False,
            delay: Optional[Union[int, float]] = None
    ) -> Report[Union[A, Uncaught], Optional['ResilientCont']]:
        """
            Async - starts the chain.
            :raises TimeoutError: delay is not None and the waiting time has been exceeded.
        """
        if delay is None:
            return await self._launch(rebuild=rebuild)
        else:
            async with asyncio.timeout(delay=delay):
                return await self._launch(rebuild=rebuild)

    def to_task(self, rebuild: bool = False) -> asyncio.Task:
        """Wrap the chain starter method into a Task"""
        return asyncio.create_task(self._launch(rebuild=rebuild))


def of(value: A) -> ResilientPrime[A]:
    """
        Lazy monad for resilient async effects.
        Can accept both async and sync functions.
        Resilient means, that if there is a failure in the chain, it can return a short chain from the point of failure.
        Automatically catches errors and wraps them in an 'Uncaught' object.
        Works with bad 'Triple' and 'Uncaught' entities using the short-circuit principle.
    """
    return ResilientPrime(lambda: value)


def unit(fn: Callable[[], A]) -> ResilientPrime[A]:
    """
        Lazy monad for resilient async effects.
        Can accept both async and sync functions.
        Resilient means, that if there is a failure in the chain, it can return a short chain from the point of failure.
        Automatically catches errors and wraps them in an 'Uncaught' object.
        Works with bad 'Triple' and 'Uncaught' entities using the short-circuit principle.
    """
    return ResilientPrime(fn)


async def insist(
        resilient: Union[ResilientPrime[A], ResilientCont[A]],
        attempts: int = 1,
        delay_for_attempt: Union[int, float] = None,
        pause_between: Union[int, float] = 0
) -> Report[Union[A, Uncaught[Exc]], Optional['ResilientCont']]:
    """
        Makes 'attempts' with 'delay_for_attempt' to execute a 'resilient' chain
        with 'pause_between' intervals between them
    """
    chain, report = resilient, Report(None, resilient, None, None)

    for _ in range(attempts):
        if delay_for_attempt is None:
            report = await chain.run(rebuild=True)
        else:
            try:
                async with asyncio.timeout(delay=delay_for_attempt):
                    return await chain.run(rebuild=True)
            except TimeoutError:
                await asyncio.sleep(pause_between)
                continue

        if not report.is_ok:
            chain = report.chain_from_failure
            await asyncio.sleep(pause_between)
            continue
        return report

    return report



import unittest
import asyncio

from mafunca.triple import Left, Nothing
from mafunca.resilient import of, unit, insist
from mafunca.common.resilient_support import Report, Uncaught


async def get_five():
    return 5


async def plus_five(val):
    return val + 5


class TestResilient(unittest.IsolatedAsyncioTestCase):

    async def test_ordinary_chains1(self):
        rp = await of(10).chain(plus_five).chain(lambda v: of(v ** 2)).run()
        self.assertIsInstance(rp, Report)
        self.assertEqual(rp.chain_from_failure, None)
        self.assertEqual(rp.faulty, None)
        self.assertEqual(rp.result, 225)
        self.assertEqual(rp.last_success, None)

    async def test_ordinary_chains2(self):
        rp = await of(0).chain(lambda v: unit(get_five)).chain(plus_five).run()
        self.assertIsInstance(rp, Report)
        self.assertEqual(rp.chain_from_failure, None)
        self.assertEqual(rp.faulty, None)
        self.assertEqual(rp.result, 10)
        self.assertEqual(rp.last_success, None)

    async def test_short_circuit(self):
        rp = await of(Left('error')).chain(plus_five).run()
        self.assertEqual(rp.result.unfold(), 'error')
        self.assertEqual(rp.last_success, None)
        rp = await unit(get_five).chain(lambda _: of(Nothing())).chain(lambda v: v ** 2).run()
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(rp.last_success, None)

    async def test_catching_errors_uncaught(self):
        async def raiser():
            raise TypeError('error')

        rp = await unit(raiser).chain(lambda v: v + 1).chain(plus_five).run()
        err = rp.result
        self.assertIsInstance(err, Uncaught)
        self.assertIsInstance(err.error, TypeError)
        with self.assertRaises(TypeError):
            err.throw()

    async def test_catching_errors_caught(self):
        async def raiser():
            raise TypeError('error')

        rp = await unit(raiser).chain(plus_five).catch(lambda _: 100).run()
        self.assertEqual(rp.chain_from_failure, None)
        self.assertEqual(rp.faulty, None)
        self.assertEqual(rp.result, 100)

    async def test_catching_errors_no_effect1(self):
        async def catcher(_):
            return 100

        rp = await unit(lambda: 1).chain(plus_five).catch(catcher).chain(lambda v: v + 1).run()
        self.assertEqual(rp.result, 7)

    async def test_catching_errors_no_effect2(self):
        async def catcher(_):
            return 100

        rp = await unit(lambda: 1).chain(lambda _: Left(0)).catch(catcher).chain(plus_five).run()
        self.assertIsInstance(rp.result, Left)

    async def test_ensure_normal(self):
        g = 0

        def plus():
            nonlocal g
            g += 1

        rp = await of(10).chain(plus_five).chain(lambda v: v ** 2).ensure(plus).run()
        self.assertEqual(rp.result, 225)
        self.assertEqual(g, 1)

    async def test_ensure_uncaught(self):
        g = 0

        async def raiser():
            raise TypeError('error')

        async def plus():
            nonlocal g
            g += 1

        rp = await unit(raiser).chain(plus_five).chain(lambda v: v ** 2).ensure(plus).run()
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(g, 1)

    async def test_ensure_bad_triple(self):
        g = 0

        async def plus():
            nonlocal g
            g += 1

        rp = await unit(lambda: Nothing()).ensure(plus).chain(plus_five).run()
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(g, 1)

    async def test_nested_ordinary_chain(self):
        def inner_chain(val: int):
            return unit(lambda: val ** 2).chain(lambda v: v + 1).chain(lambda v: v + 1)

        rp = await of(5).chain(plus_five).chain(inner_chain).run()
        self.assertEqual(rp.result, 102)

    async def test_nested_short_circuit(self):
        def inner_chain(val: int):
            return of(val).chain(lambda _: Left('error')).chain(lambda v: v + 1)

        rp = await unit(get_five).chain(plus_five).chain(inner_chain).run()
        self.assertIsInstance(rp.result, Left)

    async def test_nested_uncaught_errors(self):
        async def raiser():
            raise TypeError('error')

        def inner_chain(val: int):
            return unit(raiser).chain(lambda _: val + 1).chain(lambda v: v + 1)

        rp = await unit(get_five).chain(plus_five).chain(inner_chain).run()
        self.assertIsInstance(rp.result, Uncaught)

    async def test_nested_caught_errors(self):
        async def raiser():
            raise TypeError('error')

        def inner_chain(val: int):
            return unit(raiser).chain(lambda _: val + 1).catch(lambda _: 0)

        rp = await of(5).chain(plus_five).chain(inner_chain).run()
        self.assertEqual(rp.result, 0)

    async def test_nested_caught_ensure(self):
        async def raiser():
            raise TypeError('error')

        g = 0

        def plus():
            nonlocal g
            g += 1

        def inner_chain(val: int):
            return unit(raiser).chain(lambda _: val + 1).ensure(plus)

        rp = await unit(get_five).chain(plus_five).chain(inner_chain).run()
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(g, 1)

    async def test_retry_simple(self):
        g = 0

        async def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError('error')
            return val ** 2

        resilient = of(10).chain(plus)
        rp = await resilient.run(rebuild=True)
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(rp.last_success, 10)
        rp = await rp.chain_from_failure.run(rebuild=True)
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(rp.last_success, 10)
        rp = await rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 100)
        self.assertEqual(rp.last_success, 100)

        g = -2
        rp = await insist(resilient, attempts=5)
        self.assertEqual(rp.result, 100)
        self.assertEqual(rp.last_success, 100)

    async def test_retry_simple_short_circuit(self):
        g = 0

        async def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                return Nothing()
            return val ** 2

        resilient = of(10).chain(plus)
        rp = await resilient.run(rebuild=True)
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(rp.last_success, 10)
        rp = await rp.chain_from_failure.run(rebuild=True)
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(rp.last_success, 10)
        rp = await rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 100)
        self.assertEqual(rp.last_success, 100)

        g = -2
        rp = await insist(resilient, attempts=5)
        self.assertEqual(rp.result, 100)
        self.assertEqual(rp.last_success, 100)

    async def test_retry_simple_caught(self):
        g = 0

        async def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError('error')
            return val ** 2

        resilient = of(10).chain(plus).catch(lambda _: 0)
        rp = await resilient.run(rebuild=True)
        self.assertEqual(rp.result, 0)
        self.assertEqual(rp.last_success, 0)

        g = 0
        rp = await insist(resilient, attempts=1)
        self.assertEqual(rp.result, 0)
        self.assertEqual(rp.last_success, 0)

    async def test_retry_simple_ensure(self):
        a, b = 0, 0

        async def plus(val):
            nonlocal a
            a += 1
            if a < 3:
                raise TypeError('error')
            return val ** 2

        async def ensure_plus():
            nonlocal b
            b += 1

        resilient = of(10).chain(plus).ensure(ensure_plus)
        rp = await resilient.run(rebuild=True)
        rp = await rp.chain_from_failure.run(rebuild=True)
        rp = await rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 100)
        self.assertEqual(b, 3)

        a, b = -2, 0
        rp = await insist(resilient, attempts=5)
        self.assertEqual(rp.result, 100)
        self.assertEqual(b, 5)

    async def test_retry_nested(self):
        g = 0

        async def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError('error')
            return val ** 2

        def inner(o):
            return of(o).chain(plus).chain(lambda v: unit(lambda: v + 1))

        resilient = of(5).chain(plus_five).chain(inner).chain(lambda v: v + 1)
        rp = await resilient.run(rebuild=True)
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(rp.last_success, 10)
        rp = await rp.chain_from_failure.run(rebuild=True)
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(rp.last_success, 10)
        rp = await rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 102)
        self.assertEqual(rp.last_success, 102)

        g = -2
        rp = await insist(resilient, attempts=5)
        self.assertEqual(rp.result, 102)
        self.assertEqual(rp.last_success, 102)

    async def test_retry_nested_short_circuit(self):
        g = 0

        async def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                return Nothing()
            return val ** 2

        def inner(o):
            return of(o).chain(plus).chain(lambda v: unit(lambda: v + 1))

        resilient = of(5).chain(plus_five).chain(inner).chain(lambda v: v + 1)
        rp = await resilient.run(rebuild=True)
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(rp.last_success, 10)
        rp = await rp.chain_from_failure.run(rebuild=True)
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(rp.last_success, 10)
        rp = await rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 102)
        self.assertEqual(rp.last_success, 102)

        g = -2
        rp = await insist(resilient, attempts=5)
        self.assertEqual(rp.result, 102)
        self.assertEqual(rp.last_success, 102)

    async def test_retry_nested_caught(self):
        g = 0

        async def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError('error')
            return val ** 2

        def inner(o):
            return of(o).chain(plus).catch(lambda _: 0)

        resilient = of(10).chain(plus_five).chain(inner).chain(lambda v: v + 1)
        rp = await resilient.run(rebuild=True)
        self.assertEqual(rp.result, 1)
        self.assertEqual(rp.last_success, 1)

        g = 0
        rp = await insist(resilient, attempts=1)
        self.assertEqual(rp.result, 1)
        self.assertEqual(rp.last_success, 1)

    async def test_retry_nested_ensure(self):
        a, b = 0, 0

        async def plus(val):
            nonlocal a
            a += 1
            if a < 3:
                raise TypeError('error')
            return val ** 2

        async def ensure_plus():
            nonlocal b
            b += 1

        def inner(o):
            return of(o).chain(plus).ensure(ensure_plus)

        resilient = unit(get_five).chain(plus_five).chain(inner).chain(lambda v: v + 1)
        rp = await resilient.run(rebuild=True)
        rp = await rp.chain_from_failure.run(rebuild=True)
        rp = await rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 101)
        self.assertEqual(b, 3)

        a, b = -2, 0
        rp = await insist(resilient, attempts=5)
        self.assertEqual(rp.result, 101)
        self.assertEqual(b, 5)

    async def test_delay(self):
        async def waiter(val):
            await asyncio.sleep(1)
            return val ** 2

        resilient = unit(get_five).chain(waiter)
        with self.assertRaises(TimeoutError):
            await resilient.run(delay=0.5)

        rp = await insist(resilient, attempts=2, delay_for_attempt=0.5)
        self.assertEqual(rp.chain_from_failure, resilient)

    async def test_faulty_extraction(self):
        async def failure(arg):
            return arg / 0

        async def failure_prime():
            return 10 / 0

        rp = await of(10).chain(failure).chain(lambda v: v + 1).run(rebuild=True)
        self.assertIs(rp.faulty, failure)
        self.assertEqual(rp.last_success, 10)

        rp = await unit(failure_prime).chain(failure).run(rebuild=True)
        self.assertIs(rp.faulty, failure_prime)

    async def test_catch_breaks_restored_chain(self):
        async def failure1(arg):
            return arg / 0

        async def failure2(arg):
            return arg / 0

        rp = await of(10).chain(failure1).chain(lambda v: v + 1).catch(lambda _: 0).chain(failure2).run(rebuild=True)
        self.assertIs(rp.faulty, failure2)
        self.assertEqual(rp.last_success, 0)

    async def test_single_side_effect1(self):
        kit, g = [], 0

        async def raiser(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError("it's too early")
            return val

        async def side_effect(val):
            nonlocal kit
            kit.append("side_effect")
            return val ** 2

        resilient = of(10).chain(lambda v: of(v * 2).chain(side_effect).chain(raiser)).chain(lambda v: v - 100)
        rp = await insist(resilient, 1)
        self.assertEqual(rp.last_success, 400)
        rp = await insist(rp.chain_from_failure, 2)
        self.assertEqual(rp.result, 300)
        self.assertEqual(rp.last_success, 300)
        self.assertEqual(kit, ["side_effect"])

    async def test_single_side_effect2(self):
        kit, g = [], 0

        async def side_effect_level2(val):
            nonlocal kit
            kit.append(2)
            return val + 1

        async def raiser_level2(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError("it's too early")
            return val + 1

        async def side_effect_level1(val):
            nonlocal kit
            kit.append(1)
            return val + 1

        async def raiser_level1(val):
            nonlocal g
            g += 1
            if g < 5:
                raise TypeError("it's too early")
            return val + 1

        async def side_effect_level0(val):
            nonlocal kit
            kit.append(0)
            return val ** 2

        async def level2(val):
            return unit(lambda: val + 1).chain(side_effect_level2).chain(raiser_level2).chain(lambda v: v)

        async def level1(val):
            return unit(lambda: val + 1).chain(level2).chain(side_effect_level1).chain(raiser_level1).chain(lambda v: v)

        resilient = of(0).chain(level1).chain(side_effect_level0)
        rp = await insist(resilient, 3)
        self.assertEqual(rp.last_success, 5)
        rp = await insist(rp.chain_from_failure, 10)
        self.assertEqual(rp.result, 36)
        self.assertEqual(rp.last_success, 36)
        self.assertEqual(kit, [2, 1, 0])


if __name__ == '__main__':
    unittest.main()



import unittest

from mafunca.triple import Left, Nothing
from mafunca.common.exceptions import MonadError
from mafunca.resilient_sync import of, unit, insist
from mafunca.common.resilient_support import Report, Uncaught


class TestResilientSync(unittest.TestCase):

    def test_ordinary_chains1(self):
        rp = of(10).chain(lambda v: v + 5).chain(lambda v: of(v ** 2)).run()
        self.assertIsInstance(rp, Report)
        self.assertEqual(rp.chain_from_failure, None)
        self.assertEqual(rp.faulty, None)
        self.assertEqual(rp.result, 225)
        self.assertEqual(rp.last_success, None)

    def test_ordinary_chains2(self):
        rp = of(10).chain(lambda v: unit(lambda: v + 5)).chain(lambda v: v ** 2).run()
        self.assertIsInstance(rp, Report)
        self.assertEqual(rp.chain_from_failure, None)
        self.assertEqual(rp.faulty, None)
        self.assertEqual(rp.result, 225)
        self.assertEqual(rp.last_success, None)

    def test_short_circuit(self):
        rp = of(Left('error')).chain(lambda v: unit(lambda: v + 5)).run()
        self.assertEqual(rp.result.unfold(), 'error')
        self.assertEqual(rp.last_success, None)
        rp = of(10).chain(lambda _: of(Nothing())).chain(lambda v: v ** 2).run()
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(rp.last_success, None)

    def test_catching_errors_uncaught(self):
        def raiser():
            raise TypeError('error')

        rp = unit(raiser).chain(lambda v: v + 1).chain(lambda v: v ** 2).run()
        err = rp.result
        self.assertIsInstance(err, Uncaught)
        self.assertIsInstance(err.error, TypeError)
        with self.assertRaises(TypeError):
            err.throw()

    def test_catching_errors_caught(self):
        def raiser():
            raise TypeError('error')

        rp = unit(raiser).chain(lambda v: v + 1).catch(lambda _: 100).run()
        self.assertEqual(rp.chain_from_failure, None)
        self.assertEqual(rp.faulty, None)
        self.assertEqual(rp.result, 100)

    def test_catching_errors_no_effect1(self):
        rp = unit(lambda: 1).chain(lambda v: v + 1).catch(lambda _: 100).chain(lambda v: v + 1).run()
        self.assertEqual(rp.result, 3)

    def test_catching_errors_no_effect2(self):
        rp = unit(lambda: 1).chain(lambda _: Left(0)).catch(lambda _: 100).chain(lambda v: v + 1).run()
        self.assertIsInstance(rp.result, Left)

    def test_ensure_normal(self):
        g = 0

        def plus():
            nonlocal g
            g += 1

        rp = of(10).chain(lambda v: v + 5).chain(lambda v: v ** 2).ensure(plus).run()
        self.assertEqual(rp.result, 225)
        self.assertEqual(g, 1)

    def test_ensure_uncaught(self):
        g = 0

        def raiser():
            raise TypeError('error')

        def plus():
            nonlocal g
            g += 1

        rp = unit(raiser).chain(lambda v: v + 5).chain(lambda v: v ** 2).ensure(plus).run()
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(g, 1)

    def test_ensure_bad_triple(self):
        g = 0

        def plus():
            nonlocal g
            g += 1

        rp = unit(lambda: Nothing()).ensure(plus).chain(lambda v: v + 1).run()
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(g, 1)

    def test_no_coroutine(self):
        async def coroutine1():
            return 10

        async def coroutine2(v):
            return v

        with self.assertRaises(MonadError):
            _ = unit(coroutine1)

        with self.assertRaises(MonadError):
            _ = of(10).chain(lambda _: unit(coroutine1)).run()

        with self.assertRaises(MonadError):
            _ = of(10).chain(coroutine2)

    def test_nested_ordinary_chain(self):
        def inner_chain(val: int):
            return unit(lambda: val ** 2).chain(lambda v: v + 1).chain(lambda v: v + 1)

        rp = of(5).chain(lambda v: v + 5).chain(inner_chain).run()
        self.assertEqual(rp.result, 102)

    def test_nested_short_circuit(self):

        def inner_chain(val: int):
            return of(val).chain(lambda _: Left('error')).chain(lambda v: v + 1)

        rp = of(5).chain(lambda v: v + 5).chain(inner_chain).chain(lambda v: v + 1).run()
        self.assertIsInstance(rp.result, Left)

    def test_nested_uncaught_errors(self):
        def raiser():
            raise TypeError('error')

        def inner_chain(val: int):
            return unit(raiser).chain(lambda _: val + 1).chain(lambda v: v + 1)

        rp = of(5).chain(lambda v: v + 5).chain(inner_chain).run()
        self.assertIsInstance(rp.result, Uncaught)

    def test_nested_caught_errors(self):
        def raiser():
            raise TypeError('error')

        def inner_chain(val: int):
            return unit(raiser).chain(lambda _: val + 1)

        rp = of(5).chain(lambda v: v + 5).chain(inner_chain).catch(lambda _: 0).run()
        self.assertEqual(rp.result, 0)

    def test_nested_caught_ensure(self):
        def raiser():
            raise TypeError('error')

        g = 0

        def plus():
            nonlocal g
            g += 1

        def inner_chain(val: int):
            return unit(raiser).chain(lambda _: val + 1)

        rp = of(5).chain(lambda v: v + 5).chain(inner_chain).ensure(plus).run()
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(g, 1)

    def test_retry_simple(self):
        g = 0

        def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError('error')
            return val ** 2

        resilient = of(10).chain(plus)
        rp = resilient.run(rebuild=True)
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(rp.last_success, 10)
        rp = rp.chain_from_failure.run(rebuild=True)
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(rp.last_success, 10)
        rp = rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 100)
        self.assertEqual(rp.last_success, 100)

        g = -2
        self.assertEqual(insist(resilient, attempts=5).result, 100)

    def test_retry_simple_short_circuit(self):
        g = 0

        def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                return Nothing()
            return val ** 2

        resilient = of(10).chain(plus)
        rp = resilient.run(rebuild=True)
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(rp.last_success, 10)
        rp = rp.chain_from_failure.run(rebuild=True)
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(rp.last_success, 10)
        rp = rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 100)
        self.assertEqual(rp.last_success, 100)

        g = -2
        self.assertEqual(insist(resilient, attempts=5).result, 100)

    def test_retry_simple_caught(self):
        g = 0

        def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError('error')
            return val ** 2

        resilient = of(10).chain(plus).catch(lambda _: 0)
        self.assertEqual(resilient.run(rebuild=True).result, 0)
        self.assertEqual(resilient.run(rebuild=True).last_success, 0)

        g = 0
        self.assertEqual(insist(resilient, attempts=1).result, 0)

    def test_retry_simple_ensure(self):
        a, b = 0, 0

        def plus(val):
            nonlocal a
            a += 1
            if a < 3:
                raise TypeError('error')
            return val ** 2

        def ensure_plus():
            nonlocal b
            b += 1

        resilient = of(10).chain(plus).ensure(ensure_plus)
        rp = resilient.run(rebuild=True)
        rp = rp.chain_from_failure.run(rebuild=True)
        rp = rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 100)
        self.assertEqual(b, 3)

        a, b = -2, 0
        self.assertEqual(insist(resilient, attempts=5).result, 100)
        self.assertEqual(b, 5)

    def test_retry_nested(self):
        g = 0

        def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError('error')
            return val ** 2

        def inner(o):
            return of(o).chain(plus).chain(lambda v: unit(lambda: v + 1))

        resilient = of(10).chain(lambda v: v + 10).chain(inner).chain(lambda v: v + 1)
        rp = resilient.run(rebuild=True)
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(rp.last_success, 20)
        rp = rp.chain_from_failure.run(rebuild=True)
        self.assertIsInstance(rp.result, Uncaught)
        self.assertEqual(rp.last_success, 20)
        rp = rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 402)
        self.assertEqual(rp.last_success, 402)

        g = -2
        self.assertEqual(insist(resilient, attempts=5).result, 402)

    def test_retry_nested_short_circuit(self):
        g = 0

        def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                return Nothing()
            return val ** 2

        def inner(o):
            return of(o).chain(plus).chain(lambda v: unit(lambda: v + 1))

        resilient = of(10).chain(lambda v: v + 10).chain(inner).chain(lambda v: v + 1)
        rp = resilient.run(rebuild=True)
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(rp.last_success, 20)
        rp = rp.chain_from_failure.run(rebuild=True)
        self.assertIsInstance(rp.result, Nothing)
        self.assertEqual(rp.last_success, 20)
        rp = rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 402)
        self.assertEqual(rp.last_success, 402)

        g = -2
        self.assertEqual(insist(resilient, attempts=5).result, 402)

    def test_retry_nested_caught(self):
        g = 0

        def plus(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError('error')
            return val ** 2

        def inner(o):
            return of(o).chain(plus).catch(lambda _: 0)

        resilient = of(10).chain(lambda v: v + 10).chain(inner).chain(lambda v: v + 1)
        rp = resilient.run(rebuild=True)
        self.assertEqual(rp.result, 1)
        self.assertEqual(rp.last_success, 1)

        g = 0
        self.assertEqual(insist(resilient, attempts=1).result, 1)
        self.assertEqual(insist(resilient, attempts=1).last_success, 1)

    def test_retry_nested_ensure(self):
        a, b = 0, 0

        def plus(val):
            nonlocal a
            a += 1
            if a < 3:
                raise TypeError('error')
            return val ** 2

        def ensure_plus():
            nonlocal b
            b += 1

        def inner(o):
            return of(o).chain(plus)

        resilient = of(10).chain(lambda v: v + 10).chain(inner).chain(lambda v: v + 1).ensure(ensure_plus)
        rp = resilient.run(rebuild=True)
        rp = rp.chain_from_failure.run(rebuild=True)
        rp = rp.chain_from_failure.run(rebuild=True)
        self.assertEqual(rp.result, 401)
        self.assertEqual(b, 3)

        a, b = -2, 0
        self.assertEqual(insist(resilient, attempts=5).result, 401)
        self.assertEqual(b, 5)

    def test_faulty_extraction(self):
        def failure(arg):
            return arg / 0

        def failure_prime():
            return 10 / 0

        rp = of(10).chain(failure).chain(lambda v: v + 1).run(rebuild=True)
        self.assertIs(rp.faulty, failure)
        self.assertEqual(rp.last_success, 10)

        rp = unit(failure_prime).chain(failure).run(rebuild=True)
        self.assertIs(rp.faulty, failure_prime)

    def test_catch_breaks_restored_chain(self):
        def failure1(arg):
            return arg / 0

        def failure2(arg):
            return arg / 0

        rp = of(10).chain(failure1).chain(lambda v: v + 1).catch(lambda _: 0).chain(failure2).run(rebuild=True)
        self.assertIs(rp.faulty, failure2)
        self.assertEqual(rp.last_success, 0)

    def test_single_side_effect1(self):
        kit, g = [], 0

        def raiser(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError("it's too early")
            return val

        def side_effect(val):
            nonlocal kit
            kit.append("side_effect")
            return val ** 2

        resilient = of(10).chain(lambda v: of(v * 2).chain(side_effect).chain(raiser)).chain(lambda v: v - 100)
        rp = insist(resilient, 1)
        self.assertEqual(rp.last_success, 400)
        rp = insist(rp.chain_from_failure, 2)
        self.assertEqual(rp.result, 300)
        self.assertEqual(rp.last_success, 300)
        self.assertEqual(kit, ["side_effect"])

    def test_single_side_effect2(self):
        kit, g = [], 0

        def side_effect_level2(val):
            nonlocal kit
            kit.append(2)
            return val + 1

        def raiser_level2(val):
            nonlocal g
            g += 1
            if g < 3:
                raise TypeError("it's too early")
            return val + 1

        def side_effect_level1(val):
            nonlocal kit
            kit.append(1)
            return val + 1

        def raiser_level1(val):
            nonlocal g
            g += 1
            if g < 5:
                raise TypeError("it's too early")
            return val + 1

        def side_effect_level0(val):
            nonlocal kit
            kit.append(0)
            return val ** 2

        def level2(val):
            return unit(lambda: val + 1).chain(side_effect_level2).chain(raiser_level2).chain(lambda v: v)

        def level1(val):
            return unit(lambda: val + 1).chain(level2).chain(side_effect_level1).chain(raiser_level1).chain(lambda v: v)

        resilient = of(0).chain(level1).chain(side_effect_level0)
        rp = insist(resilient, 3)
        self.assertEqual(rp.last_success, 5)
        rp = insist(rp.chain_from_failure, 10)
        self.assertEqual(rp.result, 36)
        self.assertEqual(rp.last_success, 36)
        self.assertEqual(kit, [2, 1, 0])


if __name__ == '__main__':
    unittest.main()
